
"""
This module contains a class :class:`StopWatch`. You can use stopwatches to
conveniently measure the amount of time required by some piece of code
in your program. The stopwatch will measure both, the time beginning
with the moment it was created to the moment it is added to a section
(*full time*), as well as the time of sub sections of this period
(*lap times*).

The following code snippet visualizes how to use the :class:`StopWatch`
class::

    watch = stopwatch.StopWatch()
    for i in range(10):
        compute(i)
        watch.lap("time to calculate something with parameter n=%d" % i)
    
    watch.addToSection(section)

In this snippet, ``section`` is an install of class
:class:`.vhistsection.VhistSection`.

**Warning:** On Windows, user (children) and system (children) will always be
0.
"""

import sys
import os
import time

class StopWatch(object):
    """A stopwatch class. You can use stopwatches to conventiently measure
    the time required by a peace of code.
    """
    def __init__(self):
        """Creates a new stopwatch object"""
        self.startTime = None
        self.endTime = None
        self.laps = []
    
    def _getTime(self):
        t = os.times()
        if "win32" in sys.platform:
            t = (t[0], t[1], t[2], t[3], time.clock())
        return t
    
    def lap(self, lapname):
        """Takes a lap time and starts a new lap. Lap times measure the
        time period starting with the last call of :meth:`lap` up until
        the next call. If ``lap`` has not been called before, the
        creation of the stop watch is used as the starting time.
        
        ``lapname`` is a string, which contains the name of the lap, i.e.
        ``'initialization'``, ``'iteration 1``, ``cleanup`` etc.
        """
        newLap = self._getTime()
        self.laps.append(dict(name=lapname, times=newLap))
    
    def __enter__(self):
        self.startTime = self._getTime()
        return self
    
    def __exit__(self, type, value, traceback):
        self.endTime = self._getTime()
    
    def addToSection(self, section):
        """Adds the time information to a section
        
        ``section`` must be a %%vhistsection.VhistSection object. The
        stopwatch's measurements will be added to the user-defined attributes
        of this workflow step.
        """
        total = self._formatOneLap(self.startTime, self.endTime)
        section.userAttrs["timing.total"] = total
        
        startTime = self.startTime
        for lap in self.laps:
            name = lap["name"]
            endTime = lap["times"]
            lapTime = self._formatOneLap(startTime, endTime)
            startTime = endTime
            section.userAttrs["timing.%s" % name] = lapTime
    
    def _formatOneLap(self, startTime, endTime):
        timeDiff = [e - s for s, e in zip(startTime, endTime)]
        timeStr = ("elapsed: %.2f secs, user: %.2f secs, system: %.2f secs, "
                "user (children): %.2f secs, system (children) %.2f secs")
        return timeStr % (timeDiff[4], timeDiff[0], timeDiff[1], timeDiff[2],
                timeDiff[3])

