
"""
This module contains a class %TempDir, which represents a temporary
directory.

The class %TempDir has been designed with python's ``with`` statement
in mind. This way, we can guarantee that temporary files created with
instances of this class will be deleted::

    from pyvhist.tempdir import TempDir
    
    with TempDir() as tmp:
        tmpfile1 = tmp.createTempFile("intermediate-results1")
        tmpfile2 = tmp.createTempFile("intermediate-results2")
        doSomeWork(tmpfile1, tmpfile2)
    
    # temporary files have been deleted at this point, even if "doSomeWork"
    # threw an exception
"""

import tempfile
import shutil
import os

import utils

class TempDir(object):
    """ A class, which represents a temporary directory in the file system.
    You can use this class to create temporary files, which will be
    automatically deleted.
    """
    def __init__(self):
        """ Creates a new instance of the class :class:`TempDir`. """
        datestr = utils.getDateString()
        tempSuffix = '-pyvhist-%s' % datestr
        self.__tempdir = tempfile.mkdtemp(suffix=tempSuffix)
    
    def __enter__(self):
        return self
    
    def __exit__(self, type, value, traceback):
        #print "removing tempdir %s..." % self.__tempdir
        shutil.rmtree(self.__tempdir)
    
    def getTempDir(self):
        """ Returns the path of the temporary directory, which is associated
        with this instance of :class:`TempDir`. """
        return self.__tempdir
    
    def createTempFile(self, filename):
        """ Creates a new tempfile with the name ``filename``.
        :class:`TempDir` will create subfolders for each file to make sure
        that there are no name clashes between temporary files with the same 
        name. Therefore, the following code will return two distinct temporary
        files::
            
            with TempDir() as tmp:
                tempfile1 = tmp.createTempFile("intermediate-results")
                tempfile2 = tmp.createTempFile("intermediate-results")
        """
        tmpdir = tempfile.mkdtemp(dir=self.__tempdir)
        basename = os.path.basename(filename)
        return os.path.join(tmpdir, filename)

