"""
This module provides functions which help to document the state
of a *version control system* at the time, the VHIST file has been created.
If you store your sourcecode in a repository created with one of the version 
control systems ``git`` or ``svn``, you can use these routines to document
which version of the source code was used during your program's run.

Each routine in this module will document the revision number or id, the
status of the working copy as well as uncommitted changes.

::

    from pyvhist.vhistsection import VhistSection
    from pyvhist.tempdir import TempDir
    from pyvhist.vcstools import saveGitInfo
    
    section = VhistSection()
    temp = TempDir()
    saveGitInfo("/path/to/my/sourcecode/", section, temp)
"""

import sys
import os
import re

from runprocess import runProcess
from utils import writeFile
import vhistfile


def findExecutable(exeName, paths):
    for path in paths:
        expandedPath = os.path.expandvars(path)
        fullPath = os.path.join(expandedPath, exeName)
        if os.path.isfile(fullPath):
            return fullPath
    raise RuntimeError("Could not find executable '%s'" % exeName)


def getGitExecutable():
    if "win32" in sys.platform:
        exeName = "git.exe"
        paths = ["%ProgramFiles%\\Git\\bin\\",
                "%ProgramFiles(x86)%\\Git\\bin\\"]
        return findExecutable(exeName, paths)
    else:
        return "git"


def findGitRootDir(subdir):
    if not os.path.exists(subdir):
        raise RuntimeError("directory '%s' does not exist" %  subdir)
    subdir = os.path.realpath(subdir)
    originalDir = subdir
    lastSubdir = None
    
    while subdir != lastSubdir:
        if os.path.isdir(os.path.join(subdir, ".git")):
            return subdir
        lastSubdir = subdir
        subdir = os.path.dirname(subdir)
    raise RuntimeError("Could not detect root of git working directory, "
            "starting at directory '%s'" % originalDir)


def saveGitInfo(gitWorkingCopyDir, vhistSection, tempDir,
        gitExecutable=None, findWorkingCopyRoot=False,
        revisionAttribute="gitRevision", workingCopyAttribute="gitWorkingCopy",
        diffFilename="git-diff.txt",
        diffReturnAttribute="gitDiffReturnValue",
        statusFilename="git-status.txt",
        statusReturnAttribute="gitStatusReturnValue"):
    """ Retrieves information about the git working copy and documents this
        information within ``vhistSection``. The git revision is stored as
        a user-defined attribute. Both, the outputs of ``git status`` and
        ``git diff`` are stored as embedded files.
        
        ``gitWorkingCopyDir`` is the directory, which contains the git working
        copy. This directory can either be the root of the git working copy
        or a subdirectory of the working copy. If it is a sub directory, VHIST
        will only document the sub directory unless ``findWorkingCopyRoot``
        is set to ``True``.
        
        ``vhistSection`` must be a :class:`pyvhist.vhistsection.VhistSection`
        object. The git information will be stored in this section.
        
        ``tempDir`` must be a :class`pyvhist.tempdir.TempDir` object. The
        temporary files, which contain the output of ``git diff`` and
        ``git status`` are saved in this directory.
        
        ``gitExecutable`` is an optional string containing the filepath or
        name of the git executable. If not specified, pyvhist will try to
        automatically detect it.
        
        ``revisionAttribute``, ``diffReturnAttribute`` and 
        ``statusReturnAttribute`` store the names of generated user-defined
        attributes.
        
        ``diffFilename`` and ``statusFilename`` are the basenames (without a
        path) of files, which store the output of ``git status`` and
        ``git diff``.
        
        If one of the git calls fails or the function can not find the git
        executable, the function will throw an exception of the type 
        ``RuntimeError``.
    """
    
    if gitExecutable is None:
        gitExecutable = getGitExecutable()
    
    if findWorkingCopyRoot:
        gitWorkingCopyDir = findGitRootDir(gitWorkingCopyDir)
    
    (retRev, revision) = runProcess("% rev-parse HEAD", gitExecutable,
            cwd=gitWorkingCopyDir, mergeOutputs=True)
    (retStatus, status) = runProcess("% status", gitExecutable,
            cwd=gitWorkingCopyDir, mergeOutputs=True)
    (retDiff, diff) = runProcess("% diff", gitExecutable,
            cwd=gitWorkingCopyDir, mergeOutputs=True)
    
    if retRev != 0:
        raise RuntimeError("Failed to get git revision. Output was: %s" %
                revision)
    
    if retStatus != 0:
        raise RuntimeError("Failed to get git status. Output was: %s" %
                status)
    
    if retDiff != 0:
        raise RuntimeError("Failed to get git diff. Output was: %s" %
                diff)
    
    statusFilename = tempDir.createTempFile(statusFilename)
    writeFile(statusFilename, status)
    
    diffFilename = tempDir.createTempFile(diffFilename)
    writeFile(diffFilename, diff)
    
    vdiff = vhistfile.InFile(diffFilename, description="git diff output")
    vdiff.userAttrs[diffReturnAttribute] = retDiff
    
    vstatus = vhistfile.InFile(statusFilename, description="git status output")
    vstatus.userAttrs[statusReturnAttribute] = retStatus
    
    vhistSection.userAttrs[revisionAttribute] = revision.strip()
    vhistSection.userAttrs[workingCopyAttribute] = gitWorkingCopyDir
    vhistSection.files.extend([vdiff, vstatus])


def getSvnExecutable():
    if "win32" in sys.platform:
        exeName = "svn.exe"
        paths = ["%ProgramFiles%\\SlikSvn\\bin\\",
                "%ProgramFiles(x86)%\\SlikSvn\\bin\\"]
        return findExecutable(exeName, paths)
    else:
        return "svn"


def saveSvnInfo(svnWorkingCopyDir, vhistSection, tempDir, svnExecutable=None,
        revisionAttribute="svnRevision", workingCopyAttribute="gitWorkingCopy",
        diffFilename="svn-diff.txt",
        diffReturnAttribute="svnDiffReturnValue",
        statusFilename="svn-status.txt",
        statusReturnAttribute="svnStatusReturnValue"):
    """ Retrieves information about the svn working copy and documents this
        information within ``vhistSection``. The svn revision is stored as
        a user-defined attribute. Both, the outputs of ``svn status`` and
        ``git svn`` are stored as embedded files.
        
        ``svnWorkingCopyDir`` is the directory, which contains the svn working
        copy. This directory can either be the root of the svn working copy
        or a subdirectory of the working copy. If it is a sub directory, VHIST
        will only document the sub directory unless ``findWorkingCopyRoot``
        is set to ``True``.
        
        ``vhistSection`` must be a :class:`pyvhist.vhistsection.VhistSection`
        object. The svn information will be stored in this section.
        
        ``tempDir`` must be a :class`pyvhist.tempdir.TempDir` object. The
        temporary files, which contain the output of ``svn diff`` and
        ``svn status`` are saved in this directory.
        
        ``svnExecutable`` is an optional string containing the filepath or
        name of the svn executable. If not specified, pyvhist will try to
        automatically detect it.
         
        ``revisionAttribute``, ``diffReturnAttribute`` and 
        ``statusReturnAttribute`` store the names of generated user-defined
        attributes.
        
        ``diffFilename`` and ``statusFilename`` are the basenames (without a
        path) of files, which store the output of ``git status`` and
        ``git diff``.
        
        If one of the svn calls fails, the function will throw an exception of
        the type ``RuntimeError``.
    """
    if svnExecutable is None:
        svnExecutable = getSvnExecutable()
    
    (retRev, infoString) = runProcess("% info", svnExecutable,
            cwd=svnWorkingCopyDir, mergeOutputs=True)
    (retStatus, status) = runProcess("% status", svnExecutable,
            cwd=svnWorkingCopyDir, mergeOutputs=True)
    (retDiff, diff) = runProcess("% diff", svnExecutable,
            cwd=svnWorkingCopyDir, mergeOutputs=True)
    
    if retRev != 0:
        raise RuntimeException("Failed to call svn info. Output was: %s" %
                revision)
    else:
        found = re.search("Revision: (?P<rev>[0-9]+)", infoString)
        if found is not None:
            revision = found.group("rev")
        else:
            raise RuntimeError("svn info did return 0 but could not "
                "detect revision number")
    
    if retStatus != 0:
        raise RuntimeException("Failed to get svn status. Output was: %s" %
                status)
    
    if retDiff != 0:
        raise RuntimeException("Failed to get svn diff. Output was: %s" %
                diff)
    
    statusFilename = tempDir.createTempFile("svn-status.txt")
    writeFile(statusFilename, status)
    
    diffFilename = tempDir.createTempFile("svn-diff.txt")
    writeFile(diffFilename, diff)
    
    vdiff = vhistfile.InFile(diffFilename, description="svn diff output")
    vdiff.userAttrs["svnDiffReturnValue"] = retDiff
    
    vstatus = vhistfile.InFile(statusFilename, description="svn status output")
    vstatus.userAttrs["svnStatusReturnValue"] = retStatus

    vhistSection.userAttrs[revisionAttribute] = revision
    vhistSection.userAttrs[workingCopyAttribute] = svnWorkingCopyDir
    vhistSection.files.extend([vdiff, vstatus])

