"""
Module containing a file iterator, which allows to iterate over a file



This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This application is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

$Id: fileiterator.py 2040 2010-08-13 16:37:45Z ahuesgen $
"""

class FileIterator(object):
    """
    A file iterator, which can iterate linewise over a file in both directions.
    
    """
    def __init__(self, fileObj, addPositions = False, blocksize = 10000,
                 maxBlocksize = None):
        """
        Creates a new FileIterator and itializes it to start at the beginning
        of the file.
        
        @type  fileObj: file
        @param fileObj: A file or file-like object providing seek, tell and
                        read methods.
        
        @type  addPositions: bool
        @param addPositions: If True, each call to next() or prev() returns
                             a tuple containing the current line and the
                             index, at which the line starts within the file.
        
        @type  blocksize: int
        @param blocksize: The size of memory, which should be read from the
                          file at a time. This buffer may grow automatically
                          each time, a single line does not fit into this
                          buffer.
        
        @type  maxBlocksize: int
        @param maxBlocksize: The maximum blocksize, to which blocksize should
                             grow.
        
        """
        # open file
        self.file = fileObj
        self.startBlocksize = blocksize
        self.maxBlocksize = maxBlocksize
        self.addPositions = addPositions
        self.__reset()
    
    def __reset(self):
        # determine filesize
        FROM_END_OF_FILE = 2
        self.file.seek(0, FROM_END_OF_FILE)
        self.len = self.file.tell()
        # position cursor at beginning of file
        self.moveToBegin()
        # reset current blocksize
        self.blocksize = self.startBlocksize
   
    def getPosition(self):
        return self.pos
    
    
    def moveToBegin(self):
        """
        Moves the linecursor to the position before the first line.
        
        """
        self.pos = 0
        self.buffer = []
        self.bufIndex = 0
        self.halflinePossible = False
    
    def moveToEnd(self):
        """
        Moves the linecursor to the position behind the the last line.
        
        """
        self.pos = self.len
        self.buffer = []
        self.bufIndex = 0
        self.halflinePossible = False
    
    def moveToPos(self, offset):
        """
        Moves the cursor to the given position within the file measured
        in bytes. If the next call is next(), the line starting at the given
        location will be returned, in case of prev(), the line ending at the
        given location is returned. If the given position is not located at the
        beginning/end of a line, the next call to next()/prev() will return an
        incomplete line. Calling the opposite operation afterwards will return
        the complete line.
        
        @type  offset: int
        @param offset: The position in bytes, at which the file cursor is
                       placed.
        
        """
        self.pos = min(max(offset, 0), self.len)
        self.buffer = []
        self.bufIndex = 0
        self.halflinePossible = True
    
    def next(self):
        """
        Returns the next line in the file.
        
        @rtype:  str or (str, long) or None or (None, None)
        @return: Either the line or a tuple containing the line and the 
                 startposition of the line within the file. If no new next line
                 exists, None or (None, None) respectively are returned.
        
        """
        # new buffer must be read?
        if (self.buffer == []) or (self.bufIndex >= len(self.buffer)-1):
            try:
                if self.buffer != []:
                    self.pos = self.buffer[-1][1] + len(self.buffer[-1][0])
                self.__getNextBlock(1, self.len)
            except StopIteration:
                if self.addPositions:
                    return None, None
                else:
                    return None
        
        assert len(self.buffer) > 0
        self.bufIndex += 1
        line = self.buffer[self.bufIndex]
        if self.halflinePossible:
            halfline = self.buffer[0]
            del self.buffer[0]
            # special case: buffer is empty and current pos is incorrect
            if self.buffer == []:
                self.pos = halfline[1] + len(halfline[0])
            self.halflinePossible = False
            self.bufIndex -= 1
        
        if self.addPositions:
            return line
        else:
            return line[0]
    
    def prev(self):
        """
        Returns the previous line in the file.
        
        @rtype:  str or (str, long) or None or (None, None)
        @return: Either the line or a tuple containing the line and the 
                 startposition of the line within the file. If no new previous
                 line exists, None or (None, None) respectively are returned.
        
        """
        if (self.buffer == []) or (self.bufIndex <= 0):
            try:
                if self.buffer != []:
                    self.pos = self.buffer[0][1]
                self.__getNextBlock(-1, 0)
            except StopIteration:
                if self.addPositions:
                    return None, None
                else:
                    return None
        
        assert len(self.buffer) > 0
        self.bufIndex -= 1
        line = self.buffer[self.bufIndex]
        if self.halflinePossible:
            halfline = self.buffer[-1]
            del self.buffer[-1]
            # special case: buffer is empty and current pos is incorrect
            if self.buffer == []:
                self.pos = halfline[1]
            
            self.halflinePossible = False
            # no self.bufIndex +=1 since line was deleted at end of buffer
        
        if self.addPositions:
            return line
        else:
            return line[0]
    
    def readBytes(self, begin, end):
        """
        Reads the interval [begin, end) from the file. begin and end are
        automatically swapped and clammed to the files dimenstions if
        neccessary.
        
        @type  begin: int
        @param begin: The first byte, which should be read from the file.
        
        @type  end: int
        @param end: The first byte, which should not be read from the file.
        
        @rtype:  str
        @return: The section, which was read.
        
        """
        s, _, _ = self.__read(begin, end)
        return s
    
    def __read(self, begin, end):
        swapped = False
        if begin > end:
            swapped = True
            begin, end = end, begin
        
        begin = min(max(begin, 0), self.len)
        end = min(max(end, 0), self.len)
        numBytes = end - begin
        self.file.seek(begin)
        s = self.file.read(numBytes)
        if swapped:
            begin, end = end, begin
        assert len(s) == numBytes
        return s, begin, end
    
    def __increaseBlocksize(self):
        if self.blocksize == self.maxBlocksize:
            raise MemoryError, "maximum blocksize of %d bytes reached" %  \
                    self.maxBlocksize
        self.blocksize *= 2
        if self.maxBlocksize != None:
            self.blocksize = min(self.blocksize, self.maxBlocksize)
        #print "#########Blocksize: %d" % self.blocksize
    
    def __genBuffer(self, buffer, begin):
        self.buffer = []
        for line in buffer:
            self.buffer.append((line, begin))
            begin += len(line)
    
    def __getNextBlock(self, direction, endPosition):
        begin = self.pos
        end = begin + direction * self.blocksize
        
        strBuffer, begin, end = self.__read(begin, end)
        buffer = strBuffer.splitlines(True)
        self.__genBuffer(buffer, min(begin, end))
        
        if direction == -1:
            incompleteLine, newBufIndex = 0, len(self.buffer)
        else:
            incompleteLine, newBufIndex = len(self.buffer)-1, -1
        
        # we have read 0 line/bytes => file reading complete
        if len(self.buffer) == 0:
            raise StopIteration
        # only one (incomplete) line was read => resize blocksize and reread
        elif (len(self.buffer) == 1) and (end != endPosition):
            self.__increaseBlocksize()
            self.__getNextBlock(direction, endPosition)
        # we have finished reading
        elif (end == endPosition):
            self.pos = end
            self.bufIndex = newBufIndex
        # we have read more than one line (the first/last one may be incomplete)
        else:
            lineLen = len(self.buffer[incompleteLine])
            del self.buffer[incompleteLine]
            self.pos = end - direction * lineLen
            self.bufIndex = min(newBufIndex, len(self.buffer))

"""
f1 = file("fileiterator.py", "rb")
it = FileIterator(f1, False, 10)#, 60)

f = file("footest", "wb")
lst = []

it.moveToEnd()
#it.moveToBegin()
#x = it.prev()
#x = it.next()
#while x != None:
    #lst = [x] + lst
#    f.write(x)
#    x = it.next()

#f.write("".join(lst))
print repr(it.prev())
print repr(it.prev())
print repr(it.next())
print repr(it.prev())
print repr(it.prev())
print repr(it.next())

print ""
it.moveToPos(75)
print repr(it.next())
print repr(it.next())
print repr(it.prev())
#for line in it:
#    print repr(line)
"""
