"""

This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This application is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

$Id: imagehandler.py 2684 2013-04-16 14:24:21Z ahuesgen $
"""

import os
import zlib
import array
import pdfobjs
import pdftypes
import utilities

import jpeg
import png

try:
    import PIL.Image
    pilSupported = True
except ImportError, e:
    pilSupported = False


def supportsPIL():
    return pilSupported


class MetaData(object):
    def __init__(self):
        self.width = None
        self.height = None
        self.bpc = None
        self.channels = None
        self.mode = None
        self.palette = None
        self.decodeFilter = None

JPEG_EXTENSIONS = [".jpg", ".jpeg", ".jpe", ".jif", ".jfif", ".jfi"]
PNG_EXTENSIONS = [".png"]


def loadImage(filename):
    if supportsPIL():
        return ImagePIL(filename)
    else:
        root, ext = os.path.splitext(filename)
        if ext in PNG_EXTENSIONS:
            return ImagePng(filename)
        elif ext in JPEG_EXTENSIONS:
            return ImageJpeg(filename)
        else:
            raise RuntimeError("Do not know how to handle image '%s'" %
                    filename)


class ImageJpeg(object):
    def __init__(self, filename):
        metadata = jpeg.readJpegMetadata(filename)
        self.meta, self.data = self.__genImage(metadata, filename)
    
    def __genImage(self, metadata, filename):
        meta = MetaData()
        meta.width = metadata["width"]
        meta.height = metadata["height"]
        meta.bpc = metadata["precision"]
        meta.channels = metadata["components"]
        modesDict = {1: "/DeviceGray", 3: "/DeviceRGB"}
        meta.mode = modesDict[metadata["components"]]
        meta.decodeFilter = ["/DCTDecode"]
        data = utilities.readFile(filename)
        
        return meta, data
        
    def genPdfObject(self, id, gen = 0):
        obj = pdfobjs.PDFStructObject(id, gen)
        obj.set("/Type", "/XObject")
        obj.set("/Subtype", "/Image")
        obj.set("/Width", self.meta.width)
        obj.set("/Height", self.meta.height)
        obj.set("/ColorSpace", self.meta.mode)
        obj.set("/BitsPerComponent", self.meta.bpc)
        obj.set("/Filter", self.meta.decodeFilter)
        obj.set("/Length", len(self.data))
        obj.set("/Interpolate", True)
        appendix = "".join(["\nstream\n", self.data, "\nendstream\n"])
        obj.setAppendix(appendix)
        
        return [obj]



class ImagePng(object):
    def __init__(self, filename):
        data, alpha, metadata = self.loadPng(filename)
        self.meta, self.data, self.alpha = self.__genImage(metadata, data,
                alpha, filename)
    
    def __genImage(self, metadata, data, alpha, filename):
        meta = MetaData()
        meta.width = metadata["size"][0]
        meta.height = metadata["size"][1]
        meta.bpc = metadata["bitdepth"]
        meta.channels = metadata["planes"] // 2
        modesDict = {1: "/DeviceGray", 2: "/DeviceGray", 3: "/DeviceRGB",
                4: "/DeviceRGB"}
        meta.mode = modesDict[metadata["planes"]]
        meta.decodeFilter = ["/FlateDecode"]
        data = zlib.compress(data)
        if alpha is not None:
            alpha = zlib.compress(alpha)
        
        return meta, data, alpha
    
    def genPdfObject(self, id, gen = 0):
        obj = pdfobjs.PDFStructObject(id, gen)
        obj.set("/Type", "/XObject")
        obj.set("/Subtype", "/Image")
        obj.set("/Width", self.meta.width)
        obj.set("/Height", self.meta.height)
        obj.set("/ColorSpace", self.meta.mode)
        if self.meta.mode == "/Indexed":
            obj.set("/ColorSpace", [self.meta.mode, "/DeviceRGB", 255,
                    self.meta.palette])
        obj.set("/BitsPerComponent", self.meta.bpc)
        obj.set("/Filter", self.meta.decodeFilter)
        obj.set("/Length", len(self.data))
        obj.set("/Interpolate", True)
        if self.alpha != None:
            obj.set("/SMask", pdftypes.PDFRef(id+1, gen))
        appendix = "".join(["\nstream\n", self.data, "\nendstream\n"])
        obj.setAppendix(appendix)
        
        result = [obj]
        if self.alpha != None:
            obj2 = pdfobjs.PDFStructObject(id+1, gen)
            obj2.set("/Type", "/XObject")
            obj2.set("/Subtype", "/Image")
            obj2.set("/Width", self.meta.width)
            obj2.set("/Height", self.meta.height)
            obj2.set("/ColorSpace", "/DeviceGray")
            obj2.set("/BitsPerComponent", self.meta.bpc)
            obj2.set("/Filter", ["/FlateDecode"])
            obj2.set("/Length", len(self.alpha))
            obj2.set("/Interpolate", True)
            appendix = "".join(["\nstream\n", self.alpha, "\nendstream\n"])
            obj2.setAppendix(appendix)
            result.append(obj2)
        
        return result
    
    def loadRGBA(self, data):
        alpha = data[3::4]
        del data[3::4]
        alphastr = alpha.tostring()
        datastr = data.tostring()
        
        return datastr, alphastr
    
    def loadRGB(self, data):
        datastr = data.tostring()
        return datastr, None
    
    def loadLU(self, data):
        datastr = data[::1].tostring()
        alphastr = data[::2].tostring()
        return datastr, alphastr
    
    def loadL(self, data):
        return data.tostring(), None
    
    def loadPalette(self, data):
        raise NotImplementedError("loading paletted images not supported, yet")
    
    def loadPng(self, filename):
        f = open(filename, "rb")
        img = png.Reader(f)
        width, height, data, metadata = img.read_flat()
        
        alpha = metadata["alpha"]
        grayscale = metadata["greyscale"]
        planes = metadata["planes"]
        
        if alpha and not grayscale and planes > 1: # RGBA -> RGB + Alpha
            datastr, alphastr = self.loadRGBA(data)
        elif not alpha and not grayscale and planes > 1: # RGB -> RGB + None
            datastr, alphastr = self.loadRGB(data)
        elif not grayscale and planes == 1: # Palette -> Palette
            datastr, alphastr = self.loadPalette(data)
        elif not alpha and grayscale: # L -> L + None
            datastr, alphastr = self.loadLU(data)
        elif alpha and grayscale: # LU -> L + Alpha
            datastr, alphastr = self.loadL(data)
        
        return datastr, alphastr, metadata


class ImagePIL(object):
    def __init__(self, filename):
        image = PIL.Image.open(filename)
        self.meta, self.data, self.alpha = self.__genImage(image, filename)
        
    def genPalette(img):
        palette = img.palette
        if palette.mode != "RGB":
            raise RuntimeError("Color Palette of Format '%s' not supported" %
                    palette.mode)
        data = palette.getdata()[1]
        print len(data)
        paletteBytes = ["<"]
        for c in data:
            paletteBytes.append("%02x" % ord(c))
        paletteBytes.append(">")
        p = "".join(paletteBytes)
        print len(p)
        return p
    
    def __genImage(self, img, filename):
        modesDict = {"1": "/DeviceGray", "L": "/DeviceGray",
                "LA": "/DeviceGray", "P": "/Indexed", "RGB": "/DeviceRGB",
                "RGBA": "/DeviceRGB", "CMYK": "/DeviceCMYK"}
        bpcsDict = {"1": 1, "L": 8, "LA": 8, "P": 8, "RGB": 8, "RGBA": 8,
                "CMYK": 8}
        channelsDict = {"1": 1, "L": 1, "LA": 1, "P": 1, "RGB": 3, "RGBA": 3,
                "CMYK": 4}
        modesWithAlpha = ["LA", "RGBA"]
        
        meta = MetaData()
        meta.width, meta.height = img.size
        
        forceFlateForJpeg = False
        if img.mode not in modesDict:
            img = img.convert("RGB")
            forceFlateForJpeg = True
        meta.mode = modesDict[img.mode]
        
        meta.channels = channelsDict[img.mode]
        meta.bpc = bpcsDict[img.mode]
        
        if meta.mode == "/Indexed":
            meta.palette = genPalette(img)
        
        if img.format == "JPEG" and not forceFlateForJpeg:
            meta.decodeFilter = ["/DCTDecode"]
            data = utilities.readFile(filename)
            alpha = None
        else:
            meta.decodeFilter = ["/FlateDecode"]
            data = array.array("B")
            data.fromstring(img.tostring())
            alpha = None
            if img.mode in modesWithAlpha:
                # get alpha values from image and remove alpha from data
                alpha = data[meta.channels::meta.channels+1]
                del data[meta.channels::meta.channels+1]
                alpha = zlib.compress(alpha.tostring())
            data = zlib.compress(data.tostring())
        
        return meta, data, alpha
    
    def genPdfObject(self, id, gen = 0):
        obj = pdfobjs.PDFStructObject(id, gen)
        obj.set("/Type", "/XObject")
        obj.set("/Subtype", "/Image")
        obj.set("/Width", self.meta.width)
        obj.set("/Height", self.meta.height)
        obj.set("/ColorSpace", self.meta.mode)
        if self.meta.mode == "/Indexed":
            obj.set("/ColorSpace", [self.meta.mode, "/DeviceRGB", 255,
                    self.meta.palette])
        obj.set("/BitsPerComponent", self.meta.bpc)
        obj.set("/Filter", self.meta.decodeFilter)
        obj.set("/Length", len(self.data))
        obj.set("/Interpolate", True)
        if self.alpha != None:
            obj.set("/SMask", pdftypes.PDFRef(id+1, gen))
        appendix = "".join(["\nstream\n", self.data, "\nendstream\n"])
        obj.setAppendix(appendix)
        result = [obj]
        
        if self.alpha != None:
            obj2 = pdfobjs.PDFStructObject(id+1, gen)
            obj2.set("/Type", "/XObject")
            obj2.set("/Subtype", "/Image")
            obj2.set("/Width", self.meta.width)
            obj2.set("/Height", self.meta.height)
            obj2.set("/ColorSpace", "/DeviceGray")
            obj2.set("/BitsPerComponent", self.meta.bpc)
            obj2.set("/Filter", ["/FlateDecode"])
            obj2.set("/Length", len(self.alpha))
            obj2.set("/Interpolate", True)
            appendix = "".join(["\nstream\n", self.alpha, "\nendstream\n"])
            obj2.setAppendix(appendix)
            result.append(obj2)
        
        return result

