"""
Module containing unittests for lexer.py.



This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This application is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

$Id: test_lexer.py 1071 2008-05-25 18:58:08Z ahuesgen $
"""

import lexer
import unittest, verbosetestcase, re


class TestLexerError(verbosetestcase.VerboseTestCase):
    def setUp(self):
        pass
    
    def tearDown(self):
        pass
    
    def testThrow(self):
        def myThrow():
            raise lexer.LexerError, "error"
        
        # test if the correct exception type is thrown
        self.assertRaises(lexer.LexerError, myThrow)
        
        # test if the error message is transfered correctly
        try:
            raise lexer.LexerError, "error"
        except lexer.LexerError, e:
            self.assertEquals("error", str(e))

class TestLexExpr(verbosetestcase.VerboseTestCase):
    def setUp(self):
        self.testRE = re.compile("foo")
    
    def tearDown(self):
        pass
    
    def testLexExpr(self):
        def foofun(match, id):
            pass
        
        expr = lexer.LexExpr(self.testRE, True, foofun, 23)
        self.assertEquals(expr.regex, self.testRE)
        self.assertEquals(expr.endToken, True)
        self.assertEquals(expr.genFunction, foofun)
        self.assertEquals(expr.id, 23)

class TestLexer(verbosetestcase.VerboseTestCase):
    def setUp(self):
        # function for parsing tokens
        def parseToken(result, id):
            return id
        
        self.testRE1 = re.compile("foo")
        self.testRE2 = re.compile("bar")
        self.testRE3 = re.compile("foobar")
        
        self.expr1 = lexer.LexExpr(self.testRE1, False, parseToken, 1)
        self.expr2 = lexer.LexExpr(self.testRE2, False, parseToken, 2)
        self.expr3 = lexer.LexExpr(self.testRE3, False, parseToken, 3)
        self.expr4 = lexer.LexExpr(self.testRE1, True, parseToken, 4)
        
        self.myLexer = lexer.Lexer()
    
    def tearDown(self):
        pass
    
    def testSimpleLex(self):
        # add two expressions
        self.myLexer.addLexExpr(self.expr1)
        self.myLexer.addLexExpr(self.expr2)
        
        # test if the parser finds the right tokens
        mystring = "foobar"
        myresult = [1, 2]
        self.myLexer.lexString(mystring)
        result = self.myLexer.getResult()
        self.assertEquals(myresult, result)
    
    
    def testOverlappingLex1(self):
        #add three expressions
        self.myLexer.addLexExpr(self.expr1)
        self.myLexer.addLexExpr(self.expr2)
        self.myLexer.addLexExpr(self.expr3)
        
        # test if the parser finds the right tokens
        mystring = "foobar"
        myresult = [1, 2]
        self.myLexer.lexString(mystring)
        result = self.myLexer.getResult()
        self.assertEquals(myresult, result)
    
    
    def testOverlappingLex2(self):
        #add three expressions
        self.myLexer.addLexExpr(self.expr3)
        self.myLexer.addLexExpr(self.expr1)
        self.myLexer.addLexExpr(self.expr2)
        
        # test if the parser finds the right tokens
        mystring = "foobar"
        myresult = [3]
        self.myLexer.lexString(mystring)
        result = self.myLexer.getResult()
        self.assertEquals(myresult, result)
    
    
    def testTerminatedLex(self):
        #add two expressions
        self.myLexer.addLexExpr(self.expr4)
        self.myLexer.addLexExpr(self.expr2)
        
        # test if the parser finds the right tokens
        mystring = "foobar"
        myresult = [4]
        self.myLexer.lexString(mystring)
        result = self.myLexer.getResult()
        self.assertEquals(myresult, result)

if __name__ == "__main__":
    unittest.main()
