"""
Module containing unittests for pdflexer.py.



This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This application is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

$Id: test_pdflexer.py 1071 2008-05-25 18:58:08Z ahuesgen $
"""

import pdflexer, pdfutils, pdftypes
import unittest, verbosetestcase


def makeToken(text, id, pos, endpos, value = None, **kargs):
    dic = {}
    dic["TEXT"] = text
    dic["ID"] = id
    dic["POS"] = pos
    dic["ENDPOS"] = endpos
    dic["VALUE"] = value
    for key, value in kargs.iteritems():
        dic[key] = value
    return dic

class TestPDFLexer(verbosetestcase.VerboseTestCase):
    def setUp(self):
        self.pdflexer = pdflexer.PDFLexer()
    
    def tearDown(self):
        pass
    
    def testEmptyLex(self):
        self.pdflexer.lexString("")
        correctResult = []
        result = self.pdflexer.getResult()
        self.assertEquals(correctResult, result)
        
    def testDictLex(self):
        # test if << and >> are recognized
        self.pdflexer.lexString("<<>>")
        
        correctResult = [
                makeToken("<<", pdflexer.ID_DICT_BEGIN, 0, 2, {}),
                makeToken(">>", pdflexer.ID_DICT_END, 2, 4),
                        ]
        
        result = self.pdflexer.getResult()
        self.assertEquals(correctResult, result)
    
    def testArrayLex(self):
        # test if [ and ] are recognized
        self.pdflexer.lexString("[]")
        correctResult = [
                makeToken("[", pdflexer.ID_ARRAY_BEGIN, 0, 1, []),
                makeToken("]", pdflexer.ID_ARRAY_END, 1, 2)
                        ]
        
        result = self.pdflexer.getResult()
        self.assertEquals(correctResult, result)
    
    def testKeywordsLex(self):
        # test if null, true and false are recognized
        self.pdflexer.lexString("null true false")
        correctResult = [
                makeToken("null", pdflexer.ID_NULL, 0, 4, pdftypes.PDFNull()),
                makeToken("true", pdflexer.ID_BOOL, 5, 9, True),
                makeToken("false", pdflexer.ID_BOOL, 10, 15, False),
                        ]
        
        result = self.pdflexer.getResult()
        self.assertEquals(correctResult, result)
    
    def testEndObjLex(self):
        # test if endobj is recognized and lexing is stopped afterwards
        self.pdflexer.lexString("endobj false")
        correctResult = [
                makeToken("endobj", pdflexer.ID_ENDOBJ, 0, 6)
                        ]
        
        result = self.pdflexer.getResult()
        self.assertEquals(correctResult, result)
    
    def testStreamLex(self):
        # test if stream is recognized and lexing is stopped afterwards
        self.pdflexer.lexString("stream\n false")
        correctResult = [
                makeToken("stream\n", pdflexer.ID_STREAM, 0, 7)
                        ]
        
        result = self.pdflexer.getResult()
        self.assertEquals(correctResult, result)
    
    def testStartXRefLex(self):
        # test if startxref is recognized and lexing is stopped afterwards
        self.pdflexer.lexString("startxref false")
        correctResult = []
        
        result = self.pdflexer.getResult()
        self.assertEquals(correctResult, result)
    
    def testXYObjLex(self):
        # test if "x y obj" entries are recognized
        
        testData = [
                ("1 0 obj", 1, 0),
                ("10 2000 obj", 10, 2000),
                ("100   30 obj", 100, 30),
                ("000002 0003 obj", 2, 3)
                   ]
        
        for input, id, gen in testData:
            correctResult = [makeToken(input, pdflexer.ID_OBJ, 0, len(input),
                    OBJID = id, OBJGEN = gen)]
            msg = str((input, id, gen))
            self.pdflexer.lexString(input)
            result = self.pdflexer.getResult()
            self.assertEquals(correctResult, result, msg)
    
    def testReferenceLex(self):
        # test if "x y R" entries are recognized
        
        testData = [
                ("1 0 R", 1, 0, pdftypes.PDFRef(1, 0)),
                ("10 2000 R", 10, 2000, pdftypes.PDFRef(10, 2000)),
                ("100   30 R", 100, 30, pdftypes.PDFRef(100, 30)),
                ("000002 0003 R", 2, 3, pdftypes.PDFRef(2, 3))
                   ]
        
        for input, id, gen, value in testData:
            correctResult = [makeToken(input, pdflexer.ID_REF, 0, len(input),
                    value, OBJID = id, OBJGEN = gen)]
            msg = str((input, id, gen, value))
            self.pdflexer.lexString(input)
            result = self.pdflexer.getResult()
            self.assertEquals(correctResult, result, msg)
    
    def testLitStringLex(self):
        # test if various literal strings are recognized
        testData = ["()", "(foo)", "(\\()", "(\\))", "(\\\\)", "(\\(\\))"]
        
        for string in testData:
            correctResult = [makeToken(string, pdflexer.ID_LITSTR, 0,
                    len(string), pdfutils.normalizePDFString(string))]
            msg = string
            self.pdflexer.lexString(string)
            result = self.pdflexer.getResult()
            self.assertEquals(correctResult, result, msg)
    
    def testHexStringLex(self):
        # test if various hexstrings are recognized
        testData = ["<>", "<AB>", "<ab>", "<12>", "<1>", "<1234567890ABCDEF>"]
        
        for string in testData:
            msg = string
            correctResult = [makeToken(string, pdflexer.ID_HEXSTR, 0,
                    len(string), pdfutils.normalizePDFString(string))]
            self.pdflexer.lexString(string)
            result = self.pdflexer.getResult()
            self.assertEquals(correctResult, result, msg)
    
    def testNameLex(self):
        # test if various name objects are recognized
        testData = [
                ("/", "/"),
                ("/foo", "/foo"),
                ("/123", "/123"),
                ("/a2", "/a2"),
                ("/#12", pdfutils.normalizePDFName("/\x12")),
                ("/#ab", pdfutils.normalizePDFName("/\xab")),
                ("/#AB", pdfutils.normalizePDFName("/\xab")),
                ("/#61", pdfutils.normalizePDFName("/a"))
                  ]
        
        for input, normalized in testData:
            msg = str((input, normalized))
            correctResult = [makeToken(input, pdflexer.ID_NAME, 0,
                    len(input), normalized)]
            self.pdflexer.lexString(input)
            result = self.pdflexer.getResult()
            self.assertEquals(correctResult, result, msg)
    
    def testNumberLex(self):
        #test if various numbers are recognized
        testData = [("0", 0), ("+0", 0), ("-0", 0), ("1234567890", 1234567890),
                ("-1234567890", -1234567890), ("+1234567890", 1234567890),
                ("0.", 0.), ("+0.", 0.), ("-0.", -0.), ("0.01", 0.01),
                ("+0.01", 0.01), ("-0.01", -0.01), ("123.456", 123.456),
                ("+123.456", 123.456), ("-123.456", -123.456)]
        
        for string, value in testData:
            msg = str((string, value))
            correctResult = [makeToken(string, pdflexer.ID_NUMBER, 0,
                    len(string), value)]
            self.pdflexer.lexString(string)
            result = self.pdflexer.getResult()
            self.assertEquals(correctResult, result, msg)

if __name__ == "__main__":
    unittest.main()
