"""
Module for creating new and extending existing vhist pdf files.

    

This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This application is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

$Id: vhistfile.py 2695 2013-04-18 14:21:11Z ahuesgen $
"""

import os, re
import fileiterator, tagutils, xref, pdfparser, pdfobjs, pdftypes, \
        vhistconsts, vhisttags

# pdf version constants
PDFFILE_HEADERS = ("%PDF-1.4", "%PDF-1.5")
PDFFILE_HEADER = PDFFILE_HEADERS[0]

# VHIST_PDF tag constants

VHIST_PDF_MARKER = "%<--! $VHIST_PDF"

#VHIST_PDF_TAG = 
def getVhistPdfTag():
    return "%%<--! $VHIST_PDF [version:%s][creator:%s]-->\n" % \
            (vhistconsts.VHIST_PDF_VERSION,
            vhistconsts.getVhistCreatorVersion())

# VHIST_SECTION tag contants
VHIST_SECTION_MARKER = "%<--! $VHIST_SECTION"

# tag parsers
VHIST_PDF_PARSER = tagutils.VHistTagParser("VHIST_PDF", ("version", "str"),
        ("creator", "str"))

VHIST_SECTION_PARSER = tagutils.VHistTagParser("VHIST_SECTION",
        ("version", "str"), ("creator", "str"), ("title", "str"),
        ("left", "int"), ("size", "int"), ("index", "int"),
        ("md5section", "md5"), ("previousmd5", "md5"),
        ("previousmarker", "int"))

class PDFFileError(Exception):
    """
    An error encountered during the creation/extension of a pdf file.
    
    """
    pass


class StreamBuffer(object):
    """
    A buffer to accumulate long strings.
    
    """
    def __init__(self):
        """
        Creates a new empty buffer.
        
        """
        self.reset()
    
    
    def reset(self):
        """
        Empties the buffer.
        
        """
        self.contentStream = []
        self.contentSize = 0
    
    
    def addContent(self, content):
        """
        Adds a string to the buffer.
        
        @type  content: str
        @param content: A string, which should be added to the end of the
                        buffer.
        
        """
        self.contentStream.append(content)
        self.contentSize += len(content)
    
    
    def getContent(self):
        """
        Returns a string containing the buffer's content.
        
        @rtype:  str
        @return: The buffer's content.
        
        """
        return "".join(self.contentStream)
    
    
    def getContentSize(self):
        """
        Returns the number of bytes stored in the buffer.
        
        @rtype:  int
        @return: The number of bytes in the stream.
        
        """
        return self.contentSize
    
    
    def joinBuffers(self, otherBuffers):
        """
        Appends the content of one buffer to the end of self.
        
        @type  otherBuffers: list of vhistfile.StreamBuffer.
        @param otherBuffers: A list of buffers, which are appended to self.
        
        """
        for buffer in otherBuffers:
            self.contentStream += buffer.contentStream
            self.contentSize += buffer.contentSize




class PDFFile(object):
    """
    A class representing a vhist pdf file. If an existing file is opened, the
    cross reference section is parsed during the creation of the pdf file.
    The class further allows to add new objects to the end of the file, extract
    existing files and append changes to the end of the file.
    
    """
    errMsg = "Error in file %s: %s"
    def __init__(self, filename, sectionTitle, readme = ""):
        """
        Creates a new PDFFile object. If the given filename points to an
        already existing file, the file's xref is parsed.
        
        @type  filename: str
        @param filename: The name of the file, which should be used.
        
        @type  readme: str
        @param readme: The readme text, which should be added to the file. The
                       text is only added if a new vhistfile is created.
        
        """
        # set file specific members
        self.filename = filename
        self.filesize = 0
        self.readme = readme
        self.sectionTitle = sectionTitle
        if os.path.isfile(filename):
            try:
                self.file = file(filename, "rb")
            except IOError, e:
                raise PDFFileError, self.errMsg % (self.filename, e)
        else:
            self.file = None
        
        # set xref members
        self.xref = xref.XRef()
        # add 0-entry to xref
        entry = xref.XRefEntry(0, 0, 65535, True, "f")
        self.xref.addEntry(entry)
        self.oldXrefPos = None
        self.xrefPositions = []
        # set vhist specific members
        self.vhistpdfVersion = "1.0" # TODO ....
        self.prevSection = {"md5": None, "pos": None, "index": 0}
        
        # set miscellaneous members
        self.newObjs = []
        
        # decide if a new file must be created or if an old one will be extended
        if self.file != None:
            try:
                # determine filesize
                STARTOFFILE = 0
                ENDOFFILE = 2
                self.file.seek(0,ENDOFFILE)
                self.filesize = self.file.tell()
                self.file.seek(0, STARTOFFILE)
                # read file header and footer
                self.vhistpdfVersion = self.__readHeader(self.file)
                self.__readFooter(self.file)
            except PDFFileError, e:
                raise PDFFileError, self.errMsg % (self.filename, e)
    
    def __readHeader(self, fileObj):
        # create file iterator
        fiter = fileiterator.FileIterator(fileObj)
        fiter.moveToBegin()
        # read first line (fileheader) and remove lineends
        header = fiter.next()
        if header == None:
            raise PDFFileError, "File seems to be empty"
        
        # does the first line of the file match with the expected pdf header?
        if header.rstrip("\r\n") not in PDFFILE_HEADERS:
            raise PDFFileError, "File is no pdf file"
        
        # read the next 5 lines and search for a VHIST_PDF tag
        lineno = 1
        MAX_LINE = 5
        line = fiter.next()
        while lineno <= MAX_LINE and line != None:
            if line.startswith(VHIST_PDF_MARKER):
                if VHIST_PDF_PARSER.match(line[1:]):
                    version = VHIST_PDF_PARSER.get("version")
                    if version != vhistconsts.VHIST_PDF_VERSION:
                        # unsuppord
                        raise PDFFileError, \
                                ("Version of vhistfile (%s) is not supported " +
                                "this version of vhistadd") % version
                    else:
                        # header is ok!
                        return version
                        break
            lineno += 1
            line = fiter.next()
        else:
            # no header found
            raise PDFFileError, "VHIST_PDF tag not found in file"
    
    def __evaluateSectionTag(self, tagParser):
        version = tagParser.get("version")
        if version != vhistconsts.VHIST_VERSION:
            return False
        else:
            self.prevSection["index"] = tagParser.get("index")
            self.prevSection["md5"]   = tagParser.get("md5section")
            self.prevSection["pos"]   = tagParser.getStartPos()
            return True
    
    def __readXRefEntry(self, line, pos, objNum, regex):
        if line == None:
            raise PDFFileError, "incomplete xref section found at %d" % pos
        else:
            found = regex.match(line)
            if found == None:
                raise PDFFileError, "broken xref section found at %d" % pos
            else:
                entryPos  = int(found.group("pos"))
                entryGen  = int(found.group("gen"))
                entryType = found.group("type")
                entry = xref.XRefEntry(entryPos, objNum, entryGen,
                        False, entryType)
                return entry
    
    def __readXRefSection(self, fiter):
        # create list for xref entries
        entries = []
        
        line, pos = fiter.next()
        if line == None:
            raise PDFFileError, "Unexpected end of file while reading xref"
        else:
            # create a regex for the xref section header
            headRegex = re.compile("(?P<start>[0-9]+) (?P<num>[0-9]+)")
            
            # parse xref section header
            line = line.rstrip("\n\r")
            found = headRegex.match(line)
            if found != None:
                start = int(found.group("start"))
                num = int(found.group("num"))
                
                # create regex to parse an xref section entry
                reStr = "(?P<pos>[0-9]{10}) (?P<gen>[0-9]{5})" + \
                        " (?P<type>[fn])(( \n)|(\r\n))"
                entryRegex = re.compile(reStr)
                
                entries = []
                # read all entries in the section
                for objNum in xrange(start, start + num):
                    line, pos = fiter.next()
                    entry = self.__readXRefEntry(line, pos, objNum, entryRegex)
                    entries.append(entry)
                return entries
            else:
                # undo last next() call
                fiter.prev()
                return None
    
    def __readPrevXRef(self, filepos, xRef, fileObj):
        # create file iterator and move it to the xref's position
        fiter = fileiterator.FileIterator(fileObj, True)
        fiter.moveToPos(filepos)
        self.xrefPositions.insert(0, filepos)
        # test for "xref" line
        line, pos = fiter.next()
        if (line != None and line.rstrip("\r\n") == "xref"):
            # read all xref sections
            entries = self.__readXRefSection(fiter)
            while entries != None:
                for entry in entries:
                    if not xRef.containsObject(entry.id):
                        xRef.addEntry(entry)
                entries = self.__readXRefSection(fiter)
        
        # search for %%EOF
        line, pos = fiter.next()
        beginPos = pos
        while line != None and line.rstrip("\n\r") != "%%EOF":
            line, pos = fiter.next()
        # find trailer block
        if line == None:
            raise PDFFileError, "Unexpected EOF while searching for pdf %%EOF"
        else:
            endPos = pos
            trailerBlock = fiter.readBytes(beginPos, endPos)
            parser = pdfparser.PDFParser()
            trailer = parser.parsePDF(trailerBlock, beginPos)
            if trailer.exists("/Prev"):
                prevPos = trailer.get("/Prev")
                self.__readPrevXRef(prevPos, xRef, fileObj)
    
    def __readFooter(self, fileObj):
        # create file iterator
        fiter = fileiterator.FileIterator(fileObj, True)
        fiter.moveToEnd()
        # read last line (eofmarker)
        eofMarker, _ = fiter.prev()
        if eofMarker == None:
            raise PDFFileError, "File seems to be empty"
        
        # check for %%EOF marker
        if eofMarker.rstrip("\r\n") != "%%EOF":
            raise PDFFileError, "File does not end with an %%EOF marker"
        
        # read xrefstartposition
        xrefPosLine, _   = fiter.prev()
        startXRefLine, _ = fiter.prev()
        
        if xrefPosLine == None or startXRefLine == None:
            raise PDFFileError, "File to short for startxref entry"
        
        if startXRefLine.rstrip("\r\n") != "startxref":
            raise PDFFileError, ("'startxref' not at last but second line " + 
                    "found %s") % startXRefLine
        
        try:
            self.oldXrefPos = int(xrefPosLine)
        except Exception, e:
            raise PDFFileError, "Expected xref position, found %s", xrefPosLine
        
        if (self.oldXrefPos < 0):
            raise PDFFileError, "xref position negative (%d)" % self.oldXrefPos
        
        # search for section tag
        line, pos = fiter.prev()
        while line != None:
            if line.startswith(VHIST_SECTION_MARKER):
                if VHIST_SECTION_PARSER.match(line[1:], pos+1):
                    if not self.__evaluateSectionTag(VHIST_SECTION_PARSER):
                        # incompatible section version => raise
                        # TODO: Do we need this? newer vhist versions should
                        # be downward compatible
                        version = VHIST_SECTION_PARSER.get("version")
                        raise PDFFileError, ("Section version (%s) not " +
                                "supported") % version
                    else:
                        # we have found a compatible section => finished
                        break
            line, pos = fiter.prev()
        else:
            # TODO: add logger warning => no section found in existing file
            pass
        
        # search for first xref at the given location
        self.__readPrevXRef(self.oldXrefPos, self.xref, fileObj)
        
    def isNewFile(self):
        """
        Returns whether or not the file points to a new pdf file.
        
        @rtype:  bool
        @return: True if the file is a new one and False otherwise.
        
        """
        return self.file == None
    
    def getNextSectionID(self):
        return self.prevSection["index"] + 1
    
    def addObject(self, obj):
        """
        Adds a new object to the pdf file.
        
        @type  obj: pdfobjs.PDFObject
        @param obj: The object, which should be added to the pdf file.
        
        """
        self.newObjs.append(obj)
    
    def getFirstFreeID(self):
        """
        Returns the first unused id in the pdf file. Everytime, a new object
        is supposed to be added to the pdf file, the number returned by this
        function should be used as the object id, except if the new object
        should override an old one.
        
        @rtype:  int
        @return: The smallest unused object id.
        
        """
        xrefIDs = self.xref.entries.keys()
        objIDs = [object.id for object in self.newObjs]
        maxID = max(xrefIDs + objIDs)
        return maxID + 1
    
    def loadObjectFromFile(self, id):
        """
        Loads, parses and returns the object with the given id from the file.
        
        @type  id: int
        @param id: The id of the object, which should be loaded from the file.
        
        @rtype:  pdfobjs.PDFStructObject
        @return: The parsed object.
        
        """
        if self.file == None:
            raise PDFFileError, self.errMsg % (self.filename,
                    "Cannot read objects from new pdf file.")
        else:
            fiter = fileiterator.FileIterator(self.file)
            begin = self.xref.entries[id].pos
            # search for object following object 'id'
            end = self.filesize + 1
            for entry in self.xref.entries.values():
                if begin < entry.pos < end:
                    end = entry.pos
            
            bytes = fiter.readBytes(begin, end)                    
            parser = pdfparser.PDFParser()
            return parser.parsePDF(bytes, begin)
    
    def appendToFile(self):
        """
        Appends all newly added added objects, a vhist section for the appended
        data and an updated xref to the end of the file. If the file does not
        exist, yet, additionally, a pdf header is added. After appending to the
        file, the PDFFile object is rendered unusable and cannot be used to add
        additional sections to the pdf file.
        
        """
        # buffers for content before and after the section tag
        outputPre = StreamBuffer()
        outputPost = StreamBuffer()
        
        # add pdf/vhist header
        if self.file == None:
            outputPre.addContent(PDFFILE_HEADER + "\n")
            outputPre.addContent("%\xFF\xFE\xFD\xFC\n")
            outputPre.addContent(getVhistPdfTag())
            outputPre.addContent("\n")
        
        if self.readme != "":
            lines = self.readme.splitlines(True)
            for line in lines:
                outputPre.addContent("% " + line)
        
        # add objects
        outputPre.addContent("\n\n")
        for obj in self.newObjs:
            pos = outputPre.getContentSize() + self.filesize
            outputPre.addContent(obj.toStr())
            entry = xref.XRefEntry(pos, obj.id, obj.gen)
            self.xref.addEntry(entry)
        
        # add section tag
        outputPre.addContent("%")
        prefixString = outputPre.getContent()
        sectionPos = outputPre.getContentSize() + self.filesize
        
        if self.prevSection["pos"] == None:
            prevMarker = None
        else:
            prevMarker = sectionPos - self.prevSection["pos"]
        section = vhisttags.SectionTag(self.sectionTitle, prefixString,
                self.prevSection["index"] + 1, self.prevSection["md5"],
                prevMarker)
        
        # add xref
        outputPost.addContent("\n\n")
        xrefPos = sectionPos + section.getTagSize() + \
            outputPost.getContentSize()
        outputPost.addContent(self.xref.toStr())
        
        # create trailer
        trailer = pdfobjs.PDFStructObject(0, 0)
        trailer.set("/Size", self.xref.getNumObjects())
        trailer.set("/Root", pdftypes.PDFRef(2, 0))
        trailer.set("/Info", pdftypes.PDFRef(1, 0))
        if self.oldXrefPos != None:
            trailer.set("/Prev", self.oldXrefPos)
        
        # add trailer
        outputPost.addContent("trailer\n")
        outputPost.addContent(trailer.getObjectBody())
        
        outputPost.addContent("startxref\n")
        outputPost.addContent(str(xrefPos) + "\n")
        outputPost.addContent("%%EOF\n")
        
        postfixString = outputPost.getContent()
        section.setSectionPostfix(postfixString)
        sectionString = section.getTag()
        
        # write changes to file
        f = file(self.filename, "ab")
        f.write(prefixString)
        f.write(sectionString)
        f.write(postfixString)
        f.close()

###############


#pdff = PDFFile("test.pdf", "%FOOOOOOOOOOOOOOOOOOBAR")
#print "opened successful"
#obj = pdfobjs.PDFStructObject(pdff.getFirstFreeID(), 0)
#obj.set("/foo", "(" + "0"*2*1024**2 + ")" )
#obj.set("/bar", "(" + "0"*2*1024**2 + ")")
#obj.set("/bar1", "(" + "0"*2*1024**2 + ")")
#obj.set("/bar2", "(" + "0"*2*1024**2 + ")")
#obj.set("/foo", 1)

#obj = pdff.loadObjectFromFile(pdff.getFirstFreeID()-1)
#print obj.toStr()
    
#pdff.addObject(obj)
#pdff.appendToFile()
#print "write successful"
#print file("test.pdf").read()
