"""
Module containing classes which describe one workflow.



This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This application is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

$Id: workflow.py 2695 2013-04-18 14:21:11Z ahuesgen $
"""

import time, os
from xmlutils import xmlify8
from utilities import assertion
import checksums
import utilities 
import namemodifier as namemod
import userinterface
import vhistconsts

LINELEN = 79


def generateOneTag(tagname, value):
    if value == None:
        return ""
    if value in (True, False):
        val = unicode(int(value))
    else:
        val = unicode(value)
    
    return "<%s>%s</%s>\n" % (tagname, xmlify8(val), tagname)


def generateOneSummaryString(string):
    # TODO: move to pdflayouter
    cleanedString = string.replace("\\", "\\\\").replace("<", "\\<").replace("_", "\\_").replace("*", "\\*").replace("=", "\\=")
    keepNewlines = True
    lines = cleanedString.splitlines(keepNewlines)
    newLines = []
    for line in lines:
        newLines.append(line)
        if line.strip() != "":
            newLines.append('\n')
    return "".join(newLines[:-1])


def fill(string, length, padding):
    return string + padding * (length - len(string))


def limitedSetter(self, key, value):
    """
    Overriden version of __setattr__. This method prevents the creation of
    new class members if the "noNewMembers" member exists.
    
    """
    if key not in self.__dict__:
        if "noNewMembers" in self.__dict__:
            raise AttributeError, "member '%s' does not exist" % key
    
    # attribute can be set
    self.__dict__[key] = value


def makeDateStr(date):
    return time.strftime("%Y-%m-%d %H:%M:%S", date)


class WorkflowError(Exception):
    pass


class WorkflowFile(object):
    """
    Contains all data related to one file used as either
    input or output in a workflow step.
    
    """
    def __init__(self):
        # filename properties
        self.fullpath = u""      # path and filename
        self.filename = u""      # filename only
        self.filepath = u""      # path only
        self.modFilename = u""   # modified filename (no path) if name is critical 
        
        # general fileproperties
        self.filetype = u""
        self.description = u""
        self.comment = u""
        self.isInfile = False
        self.isOutfile = False
        self.isSummary = False
        self.lastModified = None
        
        # user attributes
        self.userAttrs = {}
        self.userAttrsUnits = {}
        self.userAttrsOrder = []
        
        # properties related to file inclusion
        self.optional = False
        self.embedded = True
        self.autoMD5 = True
        self.previewws = False
        self.preview = False
        self.thumbnail = False
        self.thumbnailonly = False
        self.uncompressedSize = None
        self.compressedSize = None
        self.uncompressedMD5 = None
        self.compressedMD5 = None
        self.compressionMethod = u"flate"
        self.md5File = None
        self.userMD5 = None
        
        # internal properties
        self.xmlSummary = u"" # only used for xml summary file
        self.embeddedPDFObjRef = None # only for embedded files
        self.embeddedPDFFileID = u""
        
        # avoid (unintended) addition of new members
        self.noNewMembers = True
    
    def genFromUI(self, commandlineFile, category):
        """
        @type  category: str
        @param category: "in" for infile or "out" for outfile or "vhist" for
                         vhistfile
        
        """
        # shortcuts to dictionaries
        fa = commandlineFile.fileAttrs
        ff = commandlineFile.fileFlags
        ua = commandlineFile.userAttrs
        uu = commandlineFile.userAttrsUnits
        uo = commandlineFile.userAttrsOrder
        
        # initialize all members
        assertion(category in ["in", "out"], "category must be either "
                "'in' or 'out'")
        
        self.isInfile = category == "in"
        self.isOutfile = category == "out"
        
        filename = fa["filename"]
        self.setFilename(filename)
        
        self.update("description", fa["description"])
        self.update("comment", fa["comment"])
        self.update("filetype", fa["filetype"])
        self.update("md5File", fa["md5file"])
        
        self.update("optional", ff["optional"])
        self.update("embedded", ff["embed"])
        self.update("autoMD5", ff["automd5"])
        self.update("previewws", ff["previewws"])
        self.update("preview", ff["preview"])
        self.update("thumbnail", ff["thumbnail"])
        self.update("thumbnailonly", ff["thumbnailonly"])
        
        if self.previewws or self.preview or self.thumbnail or \
                self.thumbnailonly:
            self.embedded = True
        
        if ff["compress"] == True:
            self.compressionMethod = u"flate"
        # needed since ff["compress"] can also be None
        elif ff["compress"] == False:
            self.compressionMethod = u"none"
        
        # create flat copies of user attributes
        self.userAttrs = ua.copy()
        self.userAttrsUnits = uu.copy()
        self.userAttrsOrder = uo[:]
        
        # set file modification date if not set, yet
        os.stat_float_times(False)
        try:
            fileTime = time.localtime(os.path.getmtime(filename))
            self.lastModified = makeDateStr(fileTime)
        except os.error:
            pass
        
        # retrieve uncompressed md5 from file if possible
        if self.md5File != None:
            self.userMD5 = self.md5FromFile(self.md5File)
        
        # (un)compressed filesize and md5 sums are generated later
    
    def setFilename(self, filename):
        self.fullpath = filename
        self.filepath, self.filename = os.path.split(filename)
        if self.filename in ["", ".", ".."]:
            raise WorkflowError, "'%s' is no filename but a directory" % filename
        self.modFilename = namemod.modifyCriticalName(self.filename)
    
    def md5FromFile(self, filename):
        def validateMD5(s):
            if len(s) < 32:
                return False, "Too short"
            elif len(s) > 32:
                return False, "Too long"
            for c in s:
                if c not in "ABCDEFabcdef0123456789":
                    return False, "%s is not a valid character" % c
            return True, "OK"
        
        try:
            if os.path.isdir(filename):
                raise IOError(("'%s' is a directory, you need to specify "
                "a file instead") % filename)
            f = file(filename, "r")
            try:
                line = f.readline().strip().lower()
                validLine, diagnosticsStr = validateMD5(line)
                if not validLine:
                    raise WorkflowError, ("the first line of md5 file '%s' "
                            "for file '%s' does not contain a md5 sum (%s)" %
                            (self.md5File, self.fullpath, diagnosticsStr))
                else:
                    return line
            finally:
                f.close()
        except Exception, e:
            raise WorkflowError, e
    
    def setCompressedProperties(self, compressedSize, compressedMD5):
        self.update("compressedSize", compressedSize)
        self.update("compressedMD5", compressedMD5)

    def genUncompressedPropertiesFromContent(self, content):
        self.uncompressedSize = len(content)
        if self.uncompressedMD5 == None:
            self.uncompressedMD5 = checksums.md5(content).hexdigest()
    
    def genUncompressedPropertiesFromFilename(self, filename):
        try:
            if self.uncompressedSize == None:
                self.uncompressedSize = os.path.getsize(filename)
        except os.error, e:
            raise IOError, "error reading filesize of '%s': %s" % (filename, e)
        
        try:
            if self.autoMD5 or self.embedded:
                if self.uncompressedMD5 == None:
                    self.uncompressedMD5 = utilities.genMD5FromFile(filename)
        except Exception, e:
            raise
    
    def __setattr__(self, name, value):
        limitedSetter(self, name, value)
    
    def update(self, attr, value):
        if value != None:
            self.__setattr__(attr, value)
    
    """
    def setTest(self, filename = "", lastmodified = "", MD5 = "",
                isInfile = False, isOutfile = True, bEmbed = False):
        self.filename = filename
        self.lastmodified = lastmodified
        self.MD5 = MD5
        self.MD5FromFile = ""
        self.isInfile = isInfile
        self.isOutfile = isOutfile
        self.bEmbed = bEmbed
    """
    
    def setXMLSummary(self, xmlSummary = "undefined"):
        self.xmlSummary = xmlSummary
        self.description = u"Workflow summary in XML format."
        self.filetype = u"text/xml"
        self.embedded = True
        # add more properties for xml summary file
    
    def generateXMLOneFile(self):
        """
        generate XML summary for one ws file
        
        """
        szXML = ""
        szXML += generateOneTag("Filename", self.filename)
        szXML += generateOneTag("Filepath", self.fullpath)
        szXML += generateOneTag("Filetype", self.filetype)
        szXML += generateOneTag("Description", self.description)
        szXML += generateOneTag("Comment", self.comment)
        szXML += generateOneTag("LastModified", self.lastModified)
        
        # user attributes
        szXML += "<UserAttributes>\n"
        for key in self.userAttrsOrder:
            value = self.userAttrs[key]
            unit = self.userAttrsUnits.get(key, None)
            szXML += "<Attribute>\n"
            szXML += generateOneTag("Key", key)
            szXML += generateOneTag("Value", value)
            if unit is not None:
                szXML += generateOneTag("Unit", unit)
            szXML += "</Attribute>\n"
        szXML += "</UserAttributes>\n"
        
        szXML += generateOneTag("MD5File", self.md5File)
        szXML += generateOneTag("UserMD5", self.userMD5)
        szXML += generateOneTag("GenerateMD5", self.autoMD5)
        
        szXML += generateOneTag("Embedded", self.embedded)
        szXML += generateOneTag("UncompressedSize", self.uncompressedSize)
        szXML += generateOneTag("UncompressedMD5", self.uncompressedMD5)
        assertion(self.compressionMethod in ("none", "flate"),
                "compression method must be in (none, flate)")
        if self.embedded:
            szXML += generateOneTag("CompressedMethod", self.compressionMethod)
            szXML += generateOneTag("CompressedSize", self.compressedSize)
            szXML += generateOneTag("CompressedMD5", self.compressedMD5)
            szXML += generateOneTag("PreviewWS", self.previewws)
            szXML += generateOneTag("Preview", self.preview)
            szXML += generateOneTag("Thumbnail", self.thumbnail)
            szXML += generateOneTag("ThumbnailOnly", self.thumbnailonly)
        
        return szXML
    
    def generateTextUserAttrs(self):
        conv = generateOneSummaryString
        szT = ""
        if len(self.userAttrsOrder) > 0:
            header = fill("_\=\=\= User-defined arguments ", LINELEN+4, "\=")
            szT = "\n\n" + header + "_"
        for key in self.userAttrsOrder:
            val = self.userAttrs[key]
            unit = self.userAttrsUnits.get(key, None)
            
            key = conv(key)
            val = conv(val)
            
            if unit is None:
                szT += '\n\n*%s*: %s' % (key, val)
            else:
                unit = conv(unit)
                szT += '\n\n*%s [%s]*: %s' % (key, unit, val)
        return szT
    
    def generateTextForSummary(self, fid):
        conv = generateOneSummaryString
        filename = conv(self.filename)
        md5 = conv(self.uncompressedMD5)
        szT = (u"<id:%1d>*Embedded workflow summary (XML):* =%s=\n\n*Size:* "
                u"%s Bytes, *MD5:* %s\n\n\n")
        return szT % (fid, filename, self.uncompressedSize, md5)
    
    def generateTextOneFile(self, fid):
        """
        returns textual description for one file of the ws
        """
        if self.isSummary:
            return self.generateTextForSummary(fid)
        
        conv = generateOneSummaryString
        szT = u""
        filename = conv(self.filename)
        filepath = conv(self.filepath)
        filetype = conv(self.filetype)
        description = conv(self.description)
        comment = conv(self.comment)
        
        szT += u"<id:%1d>*File:* =%s=" % (fid, filename)
        
        if filepath != "":    szT += u"\n\n*Path:* %s" % filepath
        if filetype != "":    szT += u"\n\n*Filetype:* %s" % filetype
        if description != "": szT += u"*\n\nDescription:* %s" % description
        if comment != "":     szT += u"*\n\nComment:* %s" % comment
        
        szT += u"\n\n"
        
        if self.embedded:
            szT += u"*Embedded:* yes"
        else:
            szT += u"*Embedded:* no"
        
        if self.uncompressedMD5 != None:
            szT += ", *MD5:* %s" % conv(self.uncompressedMD5)
        
        if self.userMD5 != None:
            szT += u"*\n\nUser specified MD5:* %s" % conv(self.userMD5)
        
        if self.uncompressedSize != None:
            szT += u"\n\n*Size:* %d Bytes" % self.uncompressedSize
        
        if self.lastModified != None:
            if self.uncompressedSize != None:
                szT += ", "
            else:
                szT += "\n\n"
            szT += "*Last Modified:* %s" % conv(self.lastModified)

        if self.preview or self.previewws \
                or self.thumbnail or self.thumbnailonly:
            szT += "\n\n*Flag(s):* "
            first = True
            if self.previewws:                
                if not first: szT += ", "
                szT += "previewws"; first = False
            if self.preview:                
                if not first: szT += ", "
                szT += "preview"; first = False
            if self.thumbnail:                
                if not first: szT += ", "
                szT += "thumbnail"; first = False
            if self.thumbnailonly:                
                if not first: szT += ", "
                szT += "thumbnailonly"; first = False
            
        szT += self.generateTextUserAttrs()
        szT += "\n\n\n"
        return szT


class Workflow(object):
    """
    Contains all information relevant for adding one workflow step
    
    """
    def __init__(self):
        # general information
        # the filename of the generated vhistfiles.
        self.genVhistFiles = []
        self.vhistRootFile = u""
        self.rootIsOptional = False
        self.sectionIndex = 0 # section count in vhist-file, 1-based
        
        # workflow step information
        self.title = u"" # title of workflow step
        self.description = u"" # description of workflow step, usually a few lines
        self.comment = u"" # additional comment for the workflow step
        self.command = u"" # string containing the command performed during this step
        self.tool = u"" # name and version of program used in this step
        self.toolPath = u"" # path of tool (optional)
        self.creator = unicode(vhistconsts.getVhistaddVersion())
        self.host = u"" # system on which tool was run (user defined)
        self.user = u"" # current user, maybe UNIX id (optional, user defined)
        os.stat_float_times(False)
        self.timestamp = makeDateStr(time.localtime()) # date/time in UNIX standard format
        
        self.userAttrs = {}
        self.userAttrsUnits = {}
        self.userAttrsOrder = []
        self.files = [] # files used as input or output for processing step
        
        # optional information: standard set of PDF metainformation
        # these information are only embedded in newly created meta-log files
        self.PDFTitle = u""
        self.PDFAuthor = u""
        self.PDFSubject = u""
        self.PDFCreator = u""
        self.PDFKeywords = u""
        self.PDFProducer = u""
        
        # add summary file (not generated, yet) to workflow step
        self.summaryFile = WorkflowFile()
        self.summaryFile.setFilename(u"ws_summary.xml")
        self.summaryFile.isSummary = True
        self.files.append(self.summaryFile)
        
        # this prevents the (unintended) creation of new members
        # (see __setattr__)
        self.noNewMembers = True
    
    def __setattr__(self, name, value):
        limitedSetter(self, name, value)
    
    def update(self, attr, value):
        if value != None:
            self.__setattr__(attr, value)
    
    def appendFile(self, uiFile, type):
        wff = WorkflowFile()
        wff.genFromUI(uiFile, type)
        if not wff.optional or os.path.isfile(wff.fullpath):
            self.files.append(wff)
    
    def genFromUI(self, uiAttributes):
        ui = uiAttributes
        step = ui.stepAttrs
        doc = ui.docAttrs
        
        # step properties
        self.update("vhistRootFile", ui.vhistRootFile)
        self.genVhistFiles = ui.genVhistFiles[:]
        self.update("rootIsOptional", ui.rootIsOptional)

        self.update("title", step["title"])
        self.update("description", step["description"])
        self.update("comment", step["comment"])
        self.update("command", step["command"])
        self.update("tool", step["tool"])
        self.update("toolPath", step["toolpath"])
        self.update("host", step["host"])
        self.update("user", step["user"])
        
        # user defined step properties
        self.userAttrs.update(ui.stepUserAttrs)
        self.userAttrsUnits.update(ui.stepUserAttrsUnits)
        self.userAttrsOrder = ui.stepUserAttrsOrder[:]
        
        # doc properties
        self.update("PDFTitle", doc["title"])
        self.update("PDFAuthor", doc["author"])
        self.update("PDFSubject", doc["subject"])
        self.update("PDFKeywords", doc["keywords"])
        self.update("PDFProducer", doc["producer"])
        self.update("PDFCreator", doc["creator"])
        
        # add infiles
        for f in ui.infiles:
            self.appendFile(f, "in")
        
        # add outfiles
        for f in ui.outfiles:
            self.appendFile(f, "out")
    
    def toStr(self):
        print self.__dict__
        for f in self.files:
            print ""
            print f.__dict__
    
    def generateXMLSummary(self):
        """
        generate XML summary from workflow step
        
        """
        szXML = \
            "<?xml version=\"1.0\" standalone=\"yes\" ?>\n" + \
            "<vhxchange>\n" + \
            "<Workflowstep>\n"
        
        szXML += generateOneTag("VhistRootFile", self.vhistRootFile)
        szXML += generateOneTag("Timestamp", self.timestamp)
        szXML += generateOneTag("Title", self.title)
        szXML += generateOneTag("Description", self.description)
        szXML += generateOneTag("Comment", self.comment)
        szXML += generateOneTag("Command", self.command)
        szXML += generateOneTag("Tool", self.tool)
        szXML += generateOneTag("ToolPath", self.toolPath)
        szXML += generateOneTag("Host", self.host)
        szXML += generateOneTag("User", self.user)
        szXML += generateOneTag("Creator", self.creator)
        
        szXML += generateOneTag("PDFTitle", self.PDFTitle)
        szXML += generateOneTag("PDFAuthor", self.PDFAuthor)
        szXML += generateOneTag("PDFSubject", self.PDFSubject)
        szXML += generateOneTag("PDFKeywords", self.PDFKeywords)
        szXML += generateOneTag("PDFProducer", self.PDFProducer)
        szXML += generateOneTag("PDFCreator", self.PDFCreator)
        
        # user attributes
        szXML += "<UserAttributes>\n"
        for key in self.userAttrsOrder:
            value = self.userAttrs[key]
            unit = self.userAttrsUnits.get(key, None)
            szXML += "<Attribute>\n"
            szXML += generateOneTag("Key", key)
            szXML += generateOneTag("Value", value)
            if unit is not None:
                szXML += generateOneTag("Unit", unit)
            szXML += "</Attribute>\n"
        szXML += "</UserAttributes>\n"
        
        for afile in self.files:
            if afile.isInfile and afile.isOutfile:
                szXML += "<InOutFile>\n" + afile.generateXMLOneFile() + "</InOutFile>\n"
            elif afile.isInfile:
                szXML += "<InFile>\n" + afile.generateXMLOneFile() + "</InFile>\n"
            elif afile.isOutfile:
                szXML += "<OutFile>\n" + afile.generateXMLOneFile() + "</OutFile>\n"
        
        szXMLFooter = \
            "</Workflowstep>\n" + \
            "<MD5Signature>00000000000000000000000000000000</MD5Signature>\n" + \
            "</vhxchange>\n"
        
        md5signature = checksums.md5(szXML)
        md5signature.update(szXMLFooter)
        md5signature = md5signature.hexdigest()
        
        szXMLFooter = szXMLFooter.replace(
                "<MD5Signature>00000000000000000000000000000000</MD5Signature>",
                "<MD5Signature>%s</MD5Signature>" % md5signature )
        self.summaryFile.setXMLSummary(szXML + szXMLFooter)
    
    def generateTextUserAttrs(self):
        conv = generateOneSummaryString
        szT = ""
        if len(self.userAttrsOrder) > 0:
            header = fill("_\=\=\= User-defined arguments ", LINELEN+4, "\=")
            szT = "\n\n" + header + "_"
        for key in self.userAttrsOrder:
            val = self.userAttrs[key]
            unit = self.userAttrsUnits.get(key, None)
            key = conv(key)
            val = conv(val)
            if unit is None:
                szT += '\n\n*%s*: %s' % (key, val)
            else:
                unit = conv(unit)
                szT += '\n\n*%s [%s]*: %s' % (key, unit, val)
        return szT
    
    def generateTextSummary(self):
        """
        generate text summary from workflow step
        
        """
        conv = generateOneSummaryString
        title = conv(self.title)
        timestamp = conv(self.timestamp)
        tool = conv(self.tool)
        toolPath = conv(self.toolPath)
        host = conv(self.host)
        user = conv(self.user)
        description = conv(self.description)
        comment = conv(self.comment)
        command = conv(self.command)
        szT = u"*Title:* %s\n\n" % title
        szT += u"*Added:* %s" % timestamp
        if tool != "": szT += u"\n\n*Tool:* %s" % tool 
        if toolPath != "": szT += u"\n\n*Toolpath:* %s" % toolPath
        
        if host != "": szT += u"\n\n*Host:* %s" % host
        
        if user != "":
            if host != "":
                szT += ", "
            else:
                szT += "\n\n"
            szT += u"*User:* %s" % user
        if description != "": szT += u"\n\n*Description:* %s" % description
        if comment != "": szT += u"\n\n*Comment:* %s" % comment
        if command != "": szT += u"\n\n*Command:* %s" % command
        
        szT += self.generateTextUserAttrs()
        szT += "\n\n\n"
        
        def handleFileGroup(attr, files, header):
            fid = 0
            tmpStr = ""
            for afile in files:
                if afile.__dict__[attr]:
                    tmpStr += afile.generateTextOneFile(fid)
                fid += 1
            if len(tmpStr) > 0:
                tmpStr = header + tmpStr
            return tmpStr
        
        
        inputHeader = fill(u"_--- INPUT FILE(S) ", LINELEN+1, "-") + "_\n\n\n"
        outputHeader = fill(u"_--- OUTPUT FILE(S) ", LINELEN+1, "-") + "_\n\n\n"
        vhistHeader = fill(u"_--- VHIST FILE(S) ", LINELEN+1, "-") + "_\n\n\n"
        summaryHeader = fill(u"_--- SUMMARY FILE ", LINELEN+1, "-") + "_\n\n\n"
        
        szT += handleFileGroup("isInfile", self.files, inputHeader)
        szT += handleFileGroup("isOutfile", self.files, outputHeader)
        szT += handleFileGroup("isSummary", self.files, summaryHeader)
        
        return szT
