"""
Module containing classes to manage a pdf cross reference section (xref).



This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This application is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

$Id: xref.py 1071 2008-05-25 18:58:08Z ahuesgen $
"""

class XRefEntry(object):
    """
    A class storing one entry of a cross reference section of a pdf file.
    It stores the position of the object in the file (in bytes), the object
    ID, the generation, the type (n = "new", f = "free") and whether or not
    the object should be added to the crossref section.
    
    """
    def __init__(self, pos, id, gen = 0, add = True, type = "n"):
        """
        Creates a new cross reference section entry.
        
        @type  pos: int
        @param pos: The position of the object in the file (in bytes).
        
        @type  id: int
        @param id: The id of the object.
        
        @type  gen: int
        @param gen: The generation number of the object.
        
        @type  add: bool
        @param add: True if the object should be added to the cross reference
                    section.
        
        @type  type: char
        @param type: "n" for newly objects and "f" for freed objects
        
        """
        self.pos = pos
        self.id = id
        self.gen = gen
        self.add = add
        self.type = type
    
    def toStr(self):
        """
        Converts the entry to a pdf file cross reference entry string.
        
        @rtype:  str
        @return: A string containing the cross reference section entry.
        
        """
        return "%010d %05d %c \n" % (self.pos, self.gen, self.type)



class XRef(object):
    """
    A class storing the information of one pdf cross reference section.
    The information in the cross reference can be converted to a string to
    be added to the xref section in a pdf file.
    
    """
    def __init__(self):
        """
        Creates an empty cross reference section.
        
        """
        self.entries = {}
    
    def addEntry(self, entry):
        """
        Adds one entry to the cross reference section.
        
        @type  entry: XRefEntry
        @param entry: An XRefEntry which should be added to the cross
                      reference section. If entry is not an XRefEntry object,
                      a TypeError Exception is raised.
        
        """
        if isinstance(entry, XRefEntry):
            self.entries[entry.id] = entry
        else:
            raise TypeError, "entry must be of type XRefEntry"
    
    def genXRefSectionString(self, startindex, endindex):
        """
        Generates a string for one section of a cross reference. The section
        starts with a line containing the first and the number of ids in the
        section followed by the cross reference entries for all objects in the
        section. If startindex >= endindex, an empty string is returned,
        else, a section with all objects between startindex and endindex
        (excluding endindex) is returned.
        
        @type  startindex: int
        @param startindex: The id of the first object in the section.
        
        @type  endindex: int
        @param endindex: The id of the first object, which is not included in
                         the list.
        
        @rtype:  str
        @return: A string containing the the crossreference section.
        """
        if startindex < endindex:
            result = "%d %d\n" % (startindex, endindex - startindex)
            for i in xrange(startindex, endindex):
                result += self.entries[i].toStr()
            return result
        else:
            return ""
    
    def __findAddObjects(self):
        # find all ids with attribute add == True
        keys = [key for key, value in self.entries.items() if value.add == True]
        keys.sort()
        return keys
    
    def __numAddObjects(self):
        return len(self.findAddObjects())
    
    def getNumObjects(self):
        """
        Returns the number of objects in the cross reference.
        
        @rtype:  int
        @return: The number of objects in the cross reference.
         
        """
        return len(self.entries.keys())
    
    def getFirstFreeID(self):
        for i in xrange(1, len(self.entries) + 2):
            if not self.containsObject(i):
                return i
    
    def toStr(self):
        """
        Converts the cross reference into a string containing the cross
        reference section which can be included into the pdf file.
        
        @rtype:  str
        @return: A string containing the xref section.
        
        """
        # get a sorted list of all needed object ids
        keys = self.__findAddObjects()
        
        # search for continuous sequences of indices and make a string from them
        result = "xref\n"
        seqStart = curIndex = 0
        
        for key in keys:
            if key > curIndex: # sequence ends here?
                x = self.genXRefSectionString(seqStart, curIndex)
                result += x
                seqStart = curIndex = key
            curIndex += 1
        
        # add last sequence to the string
        result += self.genXRefSectionString(seqStart, curIndex)
        return result
    
    def containsObject(self, objID):
        """
        Returns whether or not the given object id is contained in the
        cross reference.
        
        @type  objID: int
        @param objID: The id of the objects, which should be tested.
        
        @rtype:  bool
        @return: True if the object is in the reference and False otherwise.
        
        """
        return self.entries.has_key(objID)
