"""
This module contains a class %VhistSection, which describes one
workflow step of a VHIST document. Each workflow step has a list of
attributes, which describe the workflow step. The following table lists all
available attributes.

==============            ===========
Attribute Name            Description
==============            ===========
title                     The title of the workflow step
description               A more elaborate description, which explains what
                          the workflowstep does. Should not be longer then one
                          or two sentences.
comment                   An additional comment. Can for example be used to 
                          describe the reason why the workflow step has been 
                          executed.
command                   The command used to execute the workflow step. By
                          default, this attribute will be set to a jsonified
                          version of ``sys.argv``.
host                      The hostname of the system, on which the workflow
                          step has been executed. By default, this attribute
                          is set to ``socket.gethostname()``.
user                      The name of the user, which executed the workflow
                          step. By default, this attribute is set to
                          ``getpass.getuser()``.
tool                      The name of the tool, which was executed during this
                          workflow step. By default, this attribute is set to
                          the version of python.
toolpath                  The path to the tool, which was used to execute this
                          workflow step. By default, it will be set to the
                          path of the python executable as well as the path
                          to python's ``dist-packages`` site.
files                     A list of files, which should be mentioned in or
                          embedded into the workflow step. See also
                          %%vhistfile.VhistFile for a description of embedded 
                          files.
previews                  A list of previews for the workflow step. Can be
                          visualizations of the input or output files,
                          generated plots etc. See :ref:`previews` for a
                          detailed explanation of workflow step previews.
userAttrs                 A dictionary containing key value pairs. This
                          dictionary can store any additional information
                          about the workflow step. The dictionary is an
                          *ordered dictionary* and will therefore remember the
                          order, in which you supplied additional user 
                          attributes. The workflow step will list the
                          user attributes in this order in the generated VHIST
                          file.
==============            ===========

Each of these attributes are mapped to one python property with the same
name. All properties are readable and writable. ``files``, ``previews``
and ``userAttrs``, however, do not allow you to replace their containers
(lists and dictionaries), but the container objects can only be manipulated.
This means that you can write

::

    section.files.append(afile)
    del section.previews[:] # clear the previews list
    section.clear()

but not

::

    section.files = []
    section.previews = dict(numIteration = 30)

If you do not set a property's value and the property has no default value,
the property will not appear in the generated VHIST file.

The following snippet shows how to use the %VhistSection class::
    
    from pyeval.vhistsection import VhistSection
    
    section = VhistSection(title="A workflow step",
            description="performs some work")
    section.comment = "Results will be used in the presentation next week"
    section.appendToVhistFile("output.vhist")
"""

from __future__ import with_statement

import sys
import os
import types
import subprocess
import socket
import getpass
import distutils.sysconfig

from typecheck import typeCheck
from external.ordereddict.ordereddict import OrderedDict
from default import default, Default
import findvhistadd
import vhistfile
import preview
import tempdir
import userattrs
import jsonify
import utils




@default
def getUser():
    """ getUser()
    Returns the name of the user, who executed this python script. """
    return getpass.getuser()

@default
def getHostname():
    """ getHostname()
    Returns the name of the computer, on which this python script was
    executed.
    """
    return socket.gethostname()


@default
def getCommandline():
    """ getCommandline()
    Returns a jsonified version of the commandline arguments in
    ``sys.argv``.
    """
    return jsonify.jsonify(sys.argv)


@default
def getTool():
    """ getTool()
    Returns the version of python used to run this script. """
    return "python %s" % sys.version.replace("\n", " ")


@default
def getToolpath():
    """ getToolpath()
    Returns the path of the python executable as well as the path of
    the path of the dist-package site as a jsonified string. """
    distSite = distutils.sysconfig.get_python_lib()
    executable = os.path.realpath(sys.executable)
    return jsonify.jsonify((executable, distSite))


def evalIfNecessary(obj):
    """ Evaluates ``obj`` if it is callable. Otherwise, returns ``obj`` """
    if callable(obj):
        return obj()
    else:
        return obj


class VhistSection(object):
    """ A class, which represents a section, you want to append to a VHIST
    file. """
    __slots__ = ["_title", "_description", "_comment", "_command", "_host",
            "_user", "_tool", "_toolpath", "_debug", "_returnCommandline",
            "_userAttrs", "_previews", "_files"]
    def __init__(self, title=None, description=None, comment=None,
            command=getCommandline, host=getHostname,
            user=getUser, tool=getTool, toolpath=getToolpath,
            debug=False, returnCommandline=False):
        """Creates a new %VhistSection object. All parameters are
        optional and each parameter is associated with the equally named
        attribute described above.""" 
        self.title = title
        self.description = description
        self.comment = comment
        
        # the values of the following attributes may be of type Default
        # => handle this appropriately
        self.command = evalIfNecessary(command)
        self.host = evalIfNecessary(host)
        self.user = evalIfNecessary(user)
        self.tool = evalIfNecessary(tool)
        self.toolpath = evalIfNecessary(toolpath)
        self._debug = False

    #def __setattr__(self, name, value):
    #    limitedSetter(self, name, value)
        self._returnCommandline = returnCommandline
        
        self._userAttrs = OrderedDict()
        self._previews = []
        self._files = []
    
    def getTitle(self):
        return self._title
    
    def setTitle(self, title):
        typeCheck("title", title, (basestring, types.NoneType))
        self._title = title
    
    def getDescription(self):
        return self._description
    
    def setDescription(self, description):
        typeCheck("description", description, (basestring, types.NoneType))
        self._description = description
    
    def getComment(self):
        return self._comment
    
    def setComment(self, comment):
        typeCheck("comment", comment, (basestring, types.NoneType))
        self._comment = comment
    
    def getCommand(self):
        return self._command
    
    def setCommand(self, command):
        typeCheck("command", command, (basestring, types.NoneType))
        self._command = command
    
    def getHost(self):
        return self._host
    
    def setHost(self, host):
        typeCheck("host", host, (basestring, types.NoneType))
        self._host = host
    
    def getUser(self):
        return self._user
    
    def setUser(self, user):
        typeCheck("user", user, (basestring, types.NoneType))
        self._user = user
    
    def getTool(self):
        return self._tool
    
    def setTool(self, tool):
        typeCheck("tool", tool, (basestring, types.NoneType))
        self._tool = tool
    
    def getToolpath(self):
        return self._toolpath
    
    def setToolpath(self, toolpath):
        typeCheck("toolpath", toolpath, (basestring, types.NoneType))
        self._toolpath = toolpath
    
    def getFiles(self):
        return self._files

    def getUserAttrs(self):
        return self._userAttrs

    def getPreviews(self):
        return self._previews

    # properties
    title = property(getTitle, setTitle,
            doc="The title of the section")
    description = property(getDescription, setDescription,
            doc="The description of the section")
    comment = property(getComment, setComment,
            doc="The comment of the section")
    command = property(getCommand, setCommand,
            doc="The command of the section")
    host = property(getHost, setHost,
            doc="The host of the section")
    user = property(getUser, setUser,
            doc="The user of the section")
    tool = property(getTool, setTool,
            doc="The tool of the section")
    toolpath = property(getToolpath, setToolpath,
            doc="The toolpath of the section")
    files = property(getFiles, None,
            doc="A list of associated files")
    previews = property(getPreviews, None,
            doc="A list of previews")
    userAttrs = property(getUserAttrs, None,
            doc="A dictionary of key-value pairs")
    
    
    def assembleCommandline(self, tempobj):
        """ Returns a list of strings, which represents the commandline
        of vhistadd (without the executable). Preview images will be created
        at a temporary directory using the ``tempobj`` object.
        
        ``tempobj`` must be of type %%tempdir.TempDir. ``assembleCommandline``
        uses it to create temporary files, such as previe images.
        """
        commandline = []
        if self.title != None:
            commandline.extend(["-s", "title", self.title])
        if self.description != None:
            commandline.extend(["-s", "description", self.description])
        if self.comment != None:
            commandline.extend(["-s", "comment", self.comment])
        if self.command != None:
            commandline.extend(["-s", "command", self.command])
        if self.host != None:
            commandline.extend(["-s", "host", self.host])
        if self.user != None:
            commandline.extend(["-s", "user", self.user])
        if self.tool != None:
            commandline.extend(["-s", "tool", self.tool])
        if self.toolpath != None:
            commandline.extend(["-s", "toolpath", self.toolpath])
        
        for key, value in self._userAttrs.items():
            strValue = jsonify.jsonify(value)
            if isinstance(value, userattrs.UserValue) and value.unit \
                    is not None:
                commandline.extend(["-W", key, strValue, value.unit])
            else:
                commandline.extend(["-U", key, strValue])
        
        for f in self.files:
            commandline.extend(f.assembleCommandline())
        
        previewfiles = self.__genPreviews(tempobj)
        
        for f in previewfiles:
            commandline.extend(f.assembleCommandline())
        
        return commandline
    
    def _runVhist(self, prefix):
        commandline = prefix
        
        with tempdir.TempDir() as tempobj:
            commandline.extend(self.assembleCommandline(tempobj))
            
            if self._returnCommandline:
                return commandline
            else:
                p = subprocess.Popen(commandline, stdout=subprocess.PIPE,
                        stderr=subprocess.STDOUT)
                
                (stdout, stderr) = p.communicate()
                print stdout
                
                return p.returncode
    
    def createVhistSection(self, filename, rootFilename = None):
        """ Adds the workflow step to the VHIST file named by
        ``filename``. If ``rootFilename`` is None, a new VHIST file
        will be created. Otherwise, the file named by ``rootFilename``
        will be copied to the location specified by ``filename`` and
        the workflow step is appended to the end of this file. If
        ``filename`` and ``rootFilename`` point to the same file,
        the section will be appended without copying. You can also specify
        a list of strings for ``filename``. In this case, n equal VHIST
        files are created.
        """
        vhistadd = findvhistadd.getCommandlinePrefix()
        if isinstance(filename, (str, unicode)):
            filename = [filename]
        
        outfileprefix = utils.flatten([["-O", f] for f in filename])
        if len(outfileprefix) == 0:
            raise ValueError("filename must be a string or a non-empty "
                    "list of strings, but was %s" % filename)
        prefix = vhistadd + outfileprefix
        if self._debug:
            prefix.append("--debug")
        if rootFilename is not None:
            prefix.extend(["-I", rootFilename])
        
        return self._runVhist(prefix)
    
    def appendToVhistFile(self, vhistfile, additionalVhistfiles=()):
        """ Appends the workflow step to an existing VHIST file with the
        file path ``vhistfile``. If the file does not exist, yet, a new
        VHIST file will be created.
        """
        vhistadd = findvhistadd.getCommandlinePrefix()
        prefix = vhistadd + ["-A", vhistfile]
        vhistoutfiles = utils.flatten(["-O", f] for f in additionalVhistfiles)
        prefix += vhistoutfiles
        if self._debug:
            prefix.append("--debug")
        
        return self._runVhist(prefix)
    
    def __genPreviewFile(self, previewobj, tempobj, nextId):
        if not isinstance(previewobj, preview.Preview):
            previewobj = preview.Preview(previewobj,
                    description="A preview image for workflow %s" % self.title,
                    basename="workflow-preview %d" % nextId)
            nextId += 1
        elif previewobj.basename is None:
            previewobj.basename = "workflow-preview %d" % nextId
            nextId += 1
        
        return previewobj.toVhistFile(tempobj), nextId
    
    def __genPreviews(self, tempobj):
        wspreviews = []
        nextId = 1
        for preview in self._previews:
            pfile, nextId = self.__genPreviewFile(preview, tempobj, nextId)
            pfile.isPreviewWs = True
            wspreviews.append(pfile)
        return wspreviews

